/*
 * This program generates a des-like pair of encrypt/decrypt routines
 * The code is sent to stdout
 * Arguments
 * #1 = # of bits of key (must be multiple of 8)
 * #2 = # of bits of data (must be multiple of 16)
 * #3 = # of steps to wind
 * #4 = random salt
 *
 * This program was written by David Ashley  dave@xdr.com
 * http://www.xdr.com/dash
 */

#define MAXDATA 1024
#define MAXKEY 1024
#define MAXWINDINGS 64

#include <stdio.h>
#include <stdlib.h>

int keylen,datalen,windings,salt;

int picks[MAXWINDINGS][3*MAXDATA>>2];


#define TABLESIZE 64
void dumptable(int num)
{
int i,j,k;
unsigned char table[TABLESIZE];
	memset(table,255,TABLESIZE);
	for(i=0;i<TABLESIZE;++i)
	{
		j=i>>2;
		for(;;)
		{
			k=rand()&TABLESIZE-1;
			if(table[k]==255)
			{
				table[k]=j;
				break;
			}
		}
	}
	printf("static unsigned char table%d[%d]=\n{\n",num,TABLESIZE);
	for(i=0;i<TABLESIZE;++i)
	{
		printf("%2d,",table[i]);
		if((i&15)==15)
			printf("\n");
	}
	printf("};\n");
}

void header(char *name)
{
	printf("void %s(unsigned char output[%d],"
		"unsigned char input[%d],\n"
		"\t\tunsigned char key[%d])\n{\n",name,
		datalen>>3,datalen>>3,keylen>>3);
	printf("unsigned char temp1[%d],temp2[%d];\n",datalen>>3,datalen>>3);
	printf("\n");
	printf("\tmemcpy(temp1,input,%d);\n",datalen>>3);
}
void tail(void)
{
	printf("}\n");
}

void makepicks(void)
{
int i,j,k,t;
int maxbit;
int numbits;
unsigned char hist[MAXDATA+MAXKEY];
int counts[64];

	memset(hist,0,sizeof(hist));
	memset(counts,0,sizeof(counts));
	maxbit=keylen+(datalen>>1);
	numbits=datalen*3>>2;
	counts[0]=maxbit;
	for(i=0;i<windings;++i)
	{
		for(j=0;j<numbits;++j)
		{
			for(;;)
			{
				k=rand()%maxbit;
				t=hist[k];
				if(!t || !counts[t-1])
				{
					picks[i][j]=k;
					++hist[k];
					--counts[t];
					++counts[t+1];
					break;
				}
			}
		}
	}
}


void permutate(int which,int right)
{
int i,j,k;
int numnibbles;
char *source;
char *dest;
int offset;
char *place;
int bit,bit7;

	numnibbles=datalen>>3;
	source=(which&1) ? "temp2" : "temp1";
	if(which==windings-1) dest="output";
	else dest=(which&1) ? "temp1" : "temp2";
	offset=right ? (datalen>>1) : 0;
	if(!right) which=windings-1-which;
	for(i=0;i<numnibbles;++i)
	{
		for(j=0;j<6;++j)
		{
			bit=picks[which][i*6+j];
			if(bit<keylen)
			{
				place="key";
			} else
			{
				place=source;
				bit=bit-keylen+offset;
			}
			if(!j)
			{
				k=i>>1;
				if(right) k+=(datalen>>4);
				if(!(i&1))
					printf("\t%s[%d]=table%d[",dest,k,which);
				else
					printf("\t%s[%d]|=table%d[",dest,k,which);
			}
			bit7=bit&7;
			printf("((%s[%d]&%d)",place,bit>>3,1<<bit7);
			if(j<bit7) printf(">>%d",bit7-j);
			else if(j>bit7) printf("<<%d",j-bit7);
			printf(")");
			if(j<5) printf(" |\n\t\t");
			else
			{
				if(!(i&1))
					printf("]<<4;\n");
				else
					printf("];\n");
			}
		}
		printf("\n");
	}
	j=datalen>>4;
	k=right ? j : 0;
	for(i=0;i<j;++i)
		printf("\t%s[%d]^=%s[%d];\n",dest,k+i,source,(k^j)+i);
	for(i=0;i<j;++i)
		printf("\t%s[%d]=%s[%d];\n",dest,(k^j)+i,source,k+i);

	printf("\n");

}

int doencrypt=1;
int dodecrypt=1;

main(int argc,char **argv)
{
int i,j,k;
int n;

	n=1;
	while(argc>n)
	{
		if(argv[n][0]!='-') break;
		if(argv[n][1]=='d') doencrypt=0;
		else if(argv[n][1]=='e') dodecrypt=0;
		else {printf("Unknown switch %c, exiting\n",argv[n][1]);exit(-1);}
		++n;
	}
	if(argc<n+4)
	{
		printf("Use: %s [options] <key bits> <data bits> <# of windings> <salt>\n",
			argv[0]);
		printf("options: -e = suppress decrypt code, -d = suppress encrypt code\n");
		exit(0);
	}
	keylen=atoi(argv[n]);
	datalen=atoi(argv[n+1]);
	windings=atoi(argv[n+2]);
	salt=atoi(argv[n+3]);
	if((keylen&7) || keylen<8 || keylen>MAXKEY)
	{
		printf("Invalid keylen %d\n",keylen);
		printf("Must be multiple of 8, >=8 && <=%d\n",MAXKEY);
		exit(1);
	}
	if((datalen&15) || datalen<16 || datalen>MAXDATA)
	{
		printf("Invalid datalen %d\n",datalen);
		printf("Must be multiple of 16, >=16 && <=%d\n",MAXDATA);
		exit(2);
	}
	if(windings<1 || windings>MAXWINDINGS)
	{
		printf("Invalid windings %d\n",windings);
		printf("Must be >=1 && <=%d\n",MAXWINDINGS);
		exit(3);
	}
	srand(salt);
	makepicks();
	printf("/*\n");
	printf(" * This program generated by deslike.c\n");
	printf(" * written by David Ashley dash@xdr.com\n");
	printf(" * http://www.xdr.com/dash\n");
	printf(" */\n\n");
	for(i=0;i<windings;++i)
		dumptable(i);

	if(doencrypt)
	{
		header("encrypt");
		for(i=0;i<windings;++i)
			permutate(i,1);
		tail();
	}
	if(dodecrypt)
	{
		header("decrypt");
		for(i=0;i<windings;++i)
			permutate(i,0);
		tail();
	}
}
